'''
Библиотека класса вариант LITE v.0.4.0.1
-------------------------------------------------------------------------------
История изменений в конце файла
OldBean, 10.04.2020
'''
#-------------------------------------------------------------------------------
'''Словарь типов устройств (кодов семеств)'''
codes = { 0xA1:['Датчик RMS', 'U'],
          0xA2:['Твердотельное реле', 'r'],
          0xA3:['PDM-контроллер', 'w'],
          0xA4:['PWM-контроллер', 'q'],
          0xA5:['Хаб 1-Wire', 'H'],
          0xA6:['1-Wire 10 bit ADC', ''],
          0xA7:['Датчик давления MPX5010DP', 'P'],
          0xA8:['Датчик расхода', 'Q'],
          0xA9:['Сервомашинка MG90S', 's'],
          0x06:['Отладочный макет', ''],
          0x28:['Датчик температуры DS18B20', 'T'],
          0xff:['Нормально разомкнутый контакт', 'Z'],
          0x00:['Нормально замкнутый контакт', 'Z'],
          0x55:['Датчик давления BMP180', 'P'],
          0x58:['Датчик давления BMP280', 'P'],
          0x59:['Датчик давления MS5611', 'P']}
#-------------------------------------------------------------------------------
'''Список единиц измерения используемых физических величин'''
units = ['', 'В', 'Вт', 'мл/час', '°C', 'мм.рт.ст.', 'мл', 'cek', 'мин', 'час']
#-------------------------------------------------------------------------------
import time, sys
import os, select, tty, termios
import io     # used to create file streams
import fcntl  # used to access I2C parameters like addresses
import smbus # С датчиками давления общаемся при помощи этой библиотеки
import math
import inspect
#-------------------------------------------------------------------------------
I2C_SLAVE = 0x703
EE_SIZE = 6
SCR_SIZE = 9
#-------------------------------------------------------------------------------
# Вспомогательные функции
#-------------------------------------------------------------------------------
def ser(coef, offset, ui):
  '''
  ser(coef, offset, ui)
  Вспомогательная функция для упаковки (сериализации) калибровочных данных (в
  6 байтов) для хранения в БД. Точно в таком же виде эти данные хранятся в
  EEPROM устройств
  '''
  ip = int(coef) & 0xffff # Целая часть коэффициента
  fp = int((coef - ip)*10000) & 0xffff # Нормированная дробная часть
  buf = b''
  buf += ip.to_bytes(2, 'little')
  buf += fp.to_bytes(2, 'little')
  buf += (offset & 0xff).to_bytes(1, 'little')
  buf += (ui & 0xff).to_bytes(1, 'little')
  return buf
#-------------------------------------------------------------------------------
def deser(buf):
  '''
  deser(buf)
  Вспомогательная функция для распаковки калибровочных данных. Результат
  возвращается в виде списка [coef, offset, ui], где coef - масштабирующий
  коэффициент, offset - смещение и ui - индекс единицы измерения
  '''
  ip = int(buf[0]) # Младший байт целой части
  ip |= (int(buf[1])) << 8 # Старший байт целой части
  fp = int(buf[2]) # Младший байт дробной части
  fp |= (int(buf[3])) << 8 # Старший байт дробной части
  return [ip + fp*0.0001, int(buf[4]), int(buf[5])]
#-------------------------------------------------------------------------------
class I2C(object):
  ''' Класс для работы с шиной i2c
  '''
  fc = 0xFE # Затычка (чтобы не мешался при анализе устройств)
  def __init__(self, addr):
    '''
    __init__(self, addr)
    Конструктор. addr - адрес на шине i2c
    '''
    self.addr = addr
    self.rf = io.open("/dev/i2c-1", "rb", buffering = 0)
    self.wf = io.open("/dev/i2c-1", "wb", buffering = 0)
    fcntl.ioctl(self.rf, I2C_SLAVE, addr)
    fcntl.ioctl(self.wf, I2C_SLAVE, addr)

  def wd(self, buf, nt = 100, slp = 0.01):
    '''
    wd(buf, nt = 100, slp = 0.01)
    Запись блока данных в устройство с проверкой готовности TWI
      buf - строка байтов
      nt - максимальное количество попыток записи
      slp - задержка между попытками в сек
    '''
    while nt > 0:
      try: # Попытка записи
        self.wf.write(buf)
        return
      except: # TWI не готов
        nt -= 1
        time.sleep(slp)
    print('Ошибка записи по адресу 0x%02X. Завершение работы'%(self.addr))
    sys.exit(1)

  def rd(self, nb, nt = 100, slp = 0.01):
    '''
    rd(nb, nt = 100, slp = 0.01)
    Чтение блока данных с проверкой готовности TWI
      nb - количество считываемых байтов
      nt - максимальное количество попыток записи
      slp - задержка между попытками в сек
    '''
    while nt > 0:
      try: # Попытка чтения
        buf = self.rf.read(nb)
        return buf
      except: # TWI не готов
        nt -= 1
        time.sleep(slp)
    print('Ошибка чтения по адресу 0x%02X. Завершение работы'%(self.addr))
    sys.exit(1)
#-------------------------------------------------------------------------------
class RMS(I2C):
  ''' Класс-обертка датчика RMS
  '''
  fc = 0xA1 # Код семейства
  def __init__(self, addr):
    '''
    __init__(self, addr)
    Конструктор. addr - адрес на шине i2c
    '''
    I2C.__init__(self, addr) # Инициализируем i2c в родительском классе
    self._calibr = self.rc() # Калибровочные данные датчиков
    self.code = 0 # Код RMS
    self.val = 0.0 # Значение RMS
    self.cutime = 0 # Время последнего обновления данных с датчика
    self.cudtmax = 5 # Период "актуальности" данных с датчика. При превышении
                     # нужно обновлять данные
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  def rc(self):
    '''
    rc()
    Читает с устройства калибровочные данные
    '''
    self.wd(b'\xb8') # Команда - считать блок данных из EEPROM
    self.wd(b'\xbe') # Команда - считать данные из рабочей области
    buf = self.rd(EE_SIZE) # Считаем калибровку
    return deser(buf) # Распаковка блока

  def wc(self, coef = 1.0, offset = 0, ui = 0):
    '''
    wc(self, coef = 1.0, offset = 0, ui = 0)
    Запись калибровочных данных в EEPROM
      coef - калибровочный коэффициент для получения физ. величины из кода
      offset - смещение кода АЦП
      ui - индекс единицы измерения физической величины в списке units
    '''
    buf = b'\x4e' # Команда записать данные в рабочую область
    buf += ser(coef, offset, ui)
    self.wd(buf) # Запись калибровок в рабочую память устройства
    self.wd(b'\x48') # Команда записать блок данных в EEPROM

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self._calibr

  @calibr.setter
  def calibr(self, value):
    '''Записывает калибровочные данные в устройство. Калибровочные данные
    должны быть указаны в формате [coef, offset, ui], где coef - калибровочный
    коэффициент, offset - смещение, ui - индекс единицы измерения. Типичные
    значения для данного датчика: [0.0734, 38, 5]'''
    self.wc(*value)
    self._calibr = value # Обновим калибровку

  def conv(self):
    '''
    conv()
    Запуск цикла преобразования датчика
    '''
    self.wd(b'\x44')

  def update(self): # Обновление данных с датчиков
    '''
    update()
    Чтение сырых данных (кода) с устройства и вычисление RMS напряжения
    '''
    # Данные уже в буфере обмена, поэтому подавать команду 0xbe не нужно
    buf = self.rd(4) # Читаем данные с датчика
    c = buf[0]
    c |= buf[1] << 8
    c |= buf[2] << 16
    c |= buf[3] << 24
    self.code = math.sqrt(c*0.0002)
    self.val = (self.code - self.calibr[1])*self.calibr[0]
    self.cutime = time.time()
  
  def cu(self):
    '''
    cu()
    Запуск преобразования с последующим обновлением
    '''
    self.conv(); self.update();
  
  @property
  def v(self):
    if time.time() - self.cutime > self.cudtmax: # Данные уже устарели
      self.cu()
    return self.val
#-------------------------------------------------------------------------------
class Relay(I2C):
  ''' Класс силового модуля релейного типа
  '''
  fc = 0xA2
  def __init__(self, addr):
    '''
    __init__(self, addr)
    Конструктор. addr - адрес на шине i2c
    '''
    I2C.__init__(self, addr) # Инициализируем i2c в родительском классе
    self.calibr = [1.0, 0, 0] # Калибровочные данные - "затычка" для реле
    self.coefi = 1.0 # Обратная величина от коэффициента
    self._code = 0
    self._val = 0.0
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  @property
  def val(self):
    return self._val

  @val.setter
  def val(self, value):
    if value < 0.5: self.code = 0
    else: self.code = 1
    self._val = value

  @property
  def code(self):
    return self._code

  @code.setter
  def code(self, value):
    buf = bytearray();
    buf.append(0x52)
    buf.append(value & 0xff)
    self.wd(buf)
    self._code = value
    self._val = (self._code - self.calibr[1])*self.calibr[0]

  def on(self): # Включение реле
    self.code = 1

  def off(self): # Выключение реле
    self.code = 0

  # Пока просто "затычки" для единообразия
  @property
  def v(self):
    return self.val

  @val.setter
  def v(self, value):
    self.val = value
#-------------------------------------------------------------------------------
class PDM(I2C):
  ''' Класс силового модуля PDM
  '''
  fc = 0xA3
  def __init__(self, addr):
    ''' Конструктор. addr - адрес на шине i2c
    '''
    I2C.__init__(self, addr) # Инициализируем i2c в родительском классе
    self._calibr = self.rc() # Калибровочные данные
    self.coefi = 1.0/self._calibr[0] # Обратная величина от коэффициента
    self.rc() # Чтение калибровочный данных и сохранение их в calibr
    self._code = 0 # Текущий уровень модуляции
    self._val = 0.0 # Текущий уровень управляемой физической величины
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  def rc(self):
    '''
    rc()
    Читает с устройства калибровочные данные
    '''
    self.wd(b'\xb8') # Команда - считать блок данных из EEPROM
    self.wd(b'\xbe') # Команда - считать данные из рабочей области
    buf = self.rd(EE_SIZE) # Считаем калибровку
    return deser(buf) # Распаковка блока

  def wc(self, coef = 1.0, offset = 0, ui = 0):
    '''
    wc(self, coef = 1.0, offset = 0, ui = 0)
    Запись калибровочных данных в EEPROM
      coef - калибровочный коэффициент для получения физ. величины из кода
      offset - смещение кода АЦП
      ui - индекс единицы измерения физической величины в списке units
    '''
    buf = b'\x4e' # Команда записать данные в рабочую область
    buf += ser(coef, offset, ui)
    self.wd(buf) # Запись калибровок в рабочую память устройства
    self.wd(b'\x48') # Команда записать блок данных в EEPROM

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self._calibr

  @calibr.setter
  def calibr(self, value):
    '''Записывает калибровочные данные в устройство. Калибровочные данные
    должны быть указаны в формате [coef, offset, ui], где coef - калибровочный
    коэффициент, offset - смещение, ui - индекс единицы измерения.'''
    self.wc(*value)
    self._calibr = value # Обновим калибровку
    self.coefi = 1/self._calibr[0]

  @property
  def val(self):
    '''
    Свойство val - значение физической величины в единицах измерения, указанных
    в калибровке
    '''
    return self._val

  @val.setter
  def val(self, value):
    '''
    Свойство val - значение физической величины в единицах измерения, указанных
    в калибровке
    '''
    c = int(100*value*self.coefi + 0.5) # Вычисляем уровень модуляции
    self.code = c # Отсылаем его устройству (контроллеру)
    self._val = value

  @property
  def code(self):
    '''
    Свойство code - уровень PDM в %
    '''
    return self._code

  @code.setter
  def code(self, value):
    '''
    Свойство code - уровень PDM в %
    '''
    if value > 100: value = 100
    if value < 0: value = 0
    buf = bytearray();
    buf.append(0x52)
    buf.append(value & 0xff)
    self.wd(buf)
    self._code = value
    self._val = 0.01*self._code*self.calibr[0]

  def on(self):
    '''
    on()
    Установка максимального уровня модуляции (100%)
    '''
    self.code = 100

  def off(self): # Выключение
    '''
    on()
    Установка минимального уровня модуляции (0%) - выключение
    '''
    self.code = 0

  # Пока просто "затычки" для единообразия
  @property
  def v(self):
    return self.val

  @val.setter
  def v(self, value):
    self.val = value
#-------------------------------------------------------------------------------
class PWM(I2C):
  ''' Класс контроллера PWM
  '''
  fc = 0xA4
  def __init__(self, addr):
    ''' Конструктор. addr - адрес на шине i2c
    '''
    I2C.__init__(self, addr) # Инициализируем i2c в родительском классе
    self._calibr = self.rc() # Калибровочные данные
    self.coefi = 1.0/self._calibr[0] # Обратная величина от коэффициента
    self._code = 0 # Текущий уровень модуляции
    '''Первый элемент _val - Текущий уровень модуляции (скорость отбора
    в случае управления клапаном отбора), второй элемент - значение интегратора
    (суммарного расхода) с момента последнего сброса интегратора'''
    self._val = 0.0
    self._t0 = time.time()
    self._Q = 0.0
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  def rc(self):
    '''
    rc()
    Читает с устройства калибровочные данные
    '''
    self.wd(b'\xb8') # Команда - считать блок данных из EEPROM
    self.wd(b'\xbe') # Команда - считать данные из рабочей области
    buf = self.rd(EE_SIZE) # Считаем калибровку
    return deser(buf) # Распаковка блока

  def wc(self, coef = 1.0, offset = 0, ui = 0):
    '''
    wc(self, coef = 1.0, offset = 0, ui = 0)
    Запись калибровочных данных в EEPROM
      coef - калибровочный коэффициент для получения физ. величины из кода
      offset - смещение кода АЦП
      ui - индекс единицы измерения физической величины в списке units
    '''
    buf = b'\x4e' # Команда записать данные в рабочую область
    buf += ser(coef, offset, ui)
    self.wd(buf) # Запись калибровок в рабочую память устройства
    self.wd(b'\x48') # Команда записать блок данных в EEPROM

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self._calibr

  @calibr.setter
  def calibr(self, value):
    '''Записывает калибровочные данные в устройство. Калибровочные данные
    должны быть указаны в формате [coef, offset, ui], где coef - калибровочный
    коэффициент, offset - смещение, ui - индекс единицы измерения.'''
    self.wc(*value)
    self._calibr = value # Обновим калибровку
    self.coefi = 1/self._calibr[0]

  @property
  def val(self):
    '''
    Свойство val - значение физической величины в единицах измерения, указанных
    в калибровке
    '''
    return self._val

  @val.setter
  def val(self, value):
    '''
    Свойство val - значение физической величины в единицах измерения, указанных
    в калибровке
    '''
    c = int(1000*value*self.coefi + 0.5) # Вычисляем код
    self.code = c # Здесь код отправляется устройству
    self._val = value

  @property
  def code(self):
    '''
    Свойство code - уровень PWM в десятых долях %
    '''
    return self._code

  @code.setter
  def code(self, value):
    '''
    Свойство code - уровень PWM в десятых долях %
    '''
    # Сначала подсчитаем расход на момент изменения скорости отбора
    now = time.time() # Текущее время
    dt = (now - self._t0)*0.000277777777778 # Время работы в часах
    self._Q += self._val*dt # Модифицируем суммарный расход
    self._t0 = now # Начинаем новый период отбора с новой скоростью
    # Теперь модифицируем скорость отбора
    if value > 1000: value = 1000
    if value < 0: value = 0
    buf = bytearray()
    buf.append(0x52) # Запись данных в регистр регулятора
    buf.append(value & 0xff)
    buf.append((value >> 8) & 0xff)
    self.wd(buf)
    self._code = value
    self._val = 0.001*self._code*self.calibr[0] # Новая скорость отбора

  def on(self):
    '''
    on()
    Установка максимального уровня модуляции (100.0%)
    '''
    self.code = 1000

  def off(self): # Выключение
    '''
    off()
    Установка минимального уровня модуляции (0%, выключение)
    '''
    self.code = 0

  # Пока просто "затычки" для единообразия
  @property
  def v(self):
    return self.val

  @val.setter
  def v(self, value):
    self.val = value
#-------------------------------------------------------------------------------
class Vol(object):
  '''
  Класс датчика расхода - отобранного объема продукта с момента последнего
  сброса (обнуления расхода). В данном релизе датчик рахода - виртуальное
  устройство и данный класс выполняет чисто интерфейсные функции. Сам расход
  подсчитывается непосредственно в объекте класса PWM, который передается
  конструктору в качестве обязательного параметра
  '''
  fc = 0xA8
  def __init__(self, pwm):
    self.pwm = pwm
    self.calibr = [1.0, 0, 6]
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  @property
  def val(self):
    '''
    Свойство - интеграл физической величины, регулируемой контроллером PWM.
    В случае клапана отбора - это полный расход через клапан с момента
    последнего сброса интегратора через сеттер
    '''
    now = time.time()
    dt = (now - self.pwm._t0)*0.000277777777778 # В часах...
    return self.pwm._Q + self.pwm._val*dt

  @val.setter
  def val(self, value):
    '''
    Сброс интегратора при любом значении value
    '''
    self.pwm._Q = 0.0
    self.pwm._t0 = time.time()
  @property
  def v(self):
    return self.val

  # Пока просто "затычки" для единообразия
  @property
  def v(self):
    return self.val

  @val.setter
  def v(self, value):
    self.val = value
#-------------------------------------------------------------------------------
class Hub(I2C):
  ''' Класс 8-канального хаба (моста) между шиной i2c и 8 шинами 1-Wire. На
      каждой шине 1-Wire должно быть только одно устройство
  '''
  fc = 0xA5
  def __init__(self, addr):
    '''
    __init__(self, addr)
    Конструктор. addr - адрес на шине i2c
    '''
    I2C.__init__(self, addr) # Инициализируем i2c в родительском классе
    self.wd(b'\x33') # Получим коды семейств подключенных к хабу датчиков
    self.fcs = [b for b in self.rd(9)[1:]]
    self.calibr = [[1.0, 0, 0]]*8 # Калибровочные данные датчиков
    self.rc() # Чтение калибровочный данных и сохранение их в calibr
    self.code = [0]*8 # Коды АЦП датчиков
    self.val = [0.0]*8 # Значения физических величин, измеряемых датчиками
    self.cutime = 0 # Время последнего обновления данных с датчика
    self.cudtmax = 5 # Период "актуальности" данных с датчика. При превышении
                     # нужно обновлять данные
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  def rc(self):
    '''
    rc()
    Чтение калибровочный данных и сохранение их в calibr
    '''
    self.wd(b'\xb8') # Команда - считать блок данных из EEPROM
    self.wd(b'\xbe') # Команда - считать данные из рабочей области
    cd = self.rd(8*SCR_SIZE) # Считаем калибровки: 8 блоков по SCR_SIZE байт

    for ch in range(8):
#      print(cd[ch*SCR_SIZE:(ch + 1)*SCR_SIZE])
      if self.fcs[ch] == 0x28: # "Затычка" для датчики DS18B20
        self.calibr[ch] = [0.0625, 0, 4]
      elif self.fcs[ch] in (0x00, 0xff): # "Затычка" для пустого канала
        self.calibr[ch] = [1.0, 0, 0]
      else:
        index = ch*SCR_SIZE + 2 # Первые 2 байта - данные. Пропускаем
        buf = cd[index:index + EE_SIZE]
        self.calibr[ch] = deser(buf)

  def wc(self, ch, coef = 1.0, offset = 0, ui = 0):
    ''' Запись калибровочных данных в EEPROM
        coef - калибровочный коэффициент для получения физ. величины из кода
        offset - смещение кода
        ui - индекс единицы измерения физической величины
    '''
    if ch < 0 or ch > 7:
      print('Неверный номер канала датчика (%d). Операция отклонена'%(ch))
      return
    buf = b'\x4e' # Команда записать данные в рабочую область
    buf += (ch & 0xff).to_bytes(1, 'little')
    buf += ser(coef, offset, ui)
    self.wd(buf) # Запись калибровок в рабочую память устройства
    self.wd(b'\x48') # Команда записать блок данных в EEPROM
    self.calibr[ch] = [coef, offset, ui] # Обновим калибровку канали ch здесь

  def conv(self): # Запуск цикла преобразования датчиков
    self.wd(b'\x44')
    self.dflag = False # Сбросим флаг готовности данных

  def update(self): # Обновление данных с датчиков
    # Данные уже в буфере обмена, поэтому подавать команду 0xbe не нужно
    buf = self.rd(16) # Читаем данные с хаба
    for ch in range(8):
      c = buf[ch*2 + 1]
      c <<= 8
      self.code[ch] = c | buf[ch*2]
      if self.fcs[ch] in (0x00, 0xff): # Контакторы, термоконтакты и т.п.
        if self.code[ch] > 1: self.val[ch] = 1
        else: self.val[ch] = 0
      else: # Другие датчики
        self.val[ch] = (self.code[ch] - self.calibr[ch][1])*self.calibr[ch][0]
    self.dflag = True
    self.cutime = time.time()

  def cu(self):
    '''
    cu()
    Запуск преобразования с последующим обновлением
    '''
    self.conv(); self.update();
#-------------------------------------------------------------------------------
class DS18B20(object):
  '''
  Класс датчиков температуры DS18B20. Эти датчики подключаются к системе через
  1-Wire хабы, объекты которых передаются в конструктор данного класса через
  параметр. Основная логика сосредоточена в хабах, поэтому данный класс
  представляет виртуальный датчик для выполнения интерфейсных функций
  '''
  fc = 0x28
  def __init__(self, hub, ch):
    '''Конструктор.
    hub - объект-хаб, к которому подключен датчик,
    ch - номер канала хаба, к которому подключен датчик
    '''
    self.hub = hub
    self.ch = ch
    self._calibr = [0.0625, 0, 4] # Калибровочные данные DS18B20
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  @property
  def val(self):
    '''
    Возвращает значение температуры, полученной после последнего вызова методов
    conv() и update() объекта hub. Для получения актуальных значений необходимо
    вызвать эти методы еще раз
    '''
    return self.hub.val[self.ch]

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self._calibr

  @property
  def v(self): # Возвращает данные с контролем их актуальности
    if time.time() - self.hub.cutime > self.hub.cudtmax: # Данные уже устарели
      self.hub.cu()
    return self.val
#-------------------------------------------------------------------------------
class MPX5010DP(object):
  '''
  Класс датчиков давления MPX5010DP. Эти датчики подключаются к системе через
  1-Wire хабы, объекты которых передаются в конструктор данного класса через
  параметр. Основная логика сосредоточена в хабах, поэтому данный класс
  представляет виртуальный датчик для выполнения интерфейсных функций
  '''
  fc = 0xA7
  def __init__(self, hub, ch):
    '''Конструктор.
    hub - объект-хаб, к которому подключен датчик,
    ch - номер канала хаба, к которому подключен датчик
    '''
    self.hub = hub
    self.ch = ch
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  @property
  def val(self):
    '''
    Возвращает значение давления, полученной после последнего вызова методов
    conv() и update() объекта hub. Для получения актуальных значений необходимо
    вызвать эти методы еще раз
    '''
    return self.hub.val[self.ch]

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self.hub.calibr[self.ch]

  @calibr.setter
  def calibr(self, value):
    '''Записывает калибровочные данные в устройство. Калибровочные данные
    должны быть указаны в формате [coef, offset, ui], где coef - калибровочный
    коэффициент, offset - смещение, ui - индекс единицы измерения. Типичные
    значения для данного датчика: [0.0734, 38, 5]'''
    self.hub.wc(self.ch, *value)
#    print('Set calibr')

  @property
  def v(self): # Возвращает данные с контролем их актуальности
    if time.time() - self.hub.cutime > self.hub.cudtmax: # Данные уже устарели
      self.hub.cu()
    return self.val
#-------------------------------------------------------------------------------
class Zcl(object):
  '''
  Класс нормально замкнутых контактов. Эти датчики подключаются к системе через
  1-Wire хабы, объекты которых передаются в конструктор данного класса через
  параметр. Основная логика сосредоточена в хабах, поэтому данный класс
  представляет виртуальный датчик для выполнения интерфейсных функций
  '''
  fc = 0x00
  def __init__(self, hub, ch):
    '''Конструктор.
    hub - объект-хаб, к которому подключен датчик,
    ch - номер канала хаба, к которому подключен датчик
    '''
    self.hub = hub
    self.ch = ch
    self._calibr = [1.0, 0, 0] # Калибровочные данные датчика
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  @property
  def val(self):
    '''
    Возвращает состояние контакта, полученной после последнего вызова методов
    conv() и update() объекта hub. Для получения актуальных значений необходимо
    вызвать эти методы еще раз
    '''
    return self.hub.val[self.ch]

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self._calibr

  @property
  def v(self): # Возвращает данные с контролем их актуальности
    if time.time() - self.hub.cutime > self.hub.cudtmax: # Данные уже устарели
      self.hub.cu()
    return self.val
#-------------------------------------------------------------------------------
class Zop(object):
  '''
  Класс нормально разомкнутых контактов. Эти датчики подключаются к системе
  через 1-Wire хабы, объекты которых передаются в конструктор данного класса
  через параметр. Основная логика сосредоточена в хабах, поэтому данный класс
  представляет виртуальный датчик для выполнения интерфейсных функций
  '''
  fc = 0xff
  def __init__(self, hub, ch):
    '''Конструктор.
    hub - объект-хаб, к которому подключен датчик,
    ch - номер канала хаба, к которому подключен датчик
    '''
    self.hub = hub
    self.ch = ch
    self._calibr = [1.0, 0, 0] # Калибровочные данные датчика
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  @property
  def val(self):
    '''
    Возвращает состояние контакта, полученной после последнего вызова методов
    conv() и update() объекта hub. Для получения актуальных значений необходимо
    вызвать эти методы еще раз
    '''
    return self.hub.val[self.ch]

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self._calibr

  @property
  def v(self): # Возвращает данные с контролем их актуальности
    if time.time() - self.hub.cutime > self.hub.cudtmax: # Данные уже устарели
      self.hub.cu()
    return self.val
#-------------------------------------------------------------------------------
class BMP180(object):
  ''' Класс датчика атмосферного давления BMP180 '''
  fc = 0x55
  def __init__(self, addr):
    '''
    __init__(self, addr)
    Конструктор. addr - адрес утройства на шине i2c.
    '''
    self.addr = addr
    self._calibr = [0.00750061683, 0, 5] # Pa -> mmHg
    self.bus = smbus.SMBus(1)
    self.temp = 0
    self.pres = 0
    # Читаем калибровочные данные
    data = self.bus.read_i2c_block_data(self.addr, 0xAA, 22)
    # Формируем 16 битные слова. Некоторые - со знаком
    self.AC1 = (data[0] << 8) + data[1]
    if self.AC1 > 32767: self.AC1 -= 65535
    self.AC2 = (data[2] << 8) + data[3]
    if self.AC2 > 32767: self.AC2 -= 65535
    self.AC3 = (data[4] << 8) + data[5]
    if self.AC3 > 32767: self.AC3 -= 65535
    self.AC4 = (data[6] << 8) + data[7]
    self.AC5 = (data[8] << 8) + data[9]
    self.AC6 = (data[10] << 8) + data[11]
    self.B1 = (data[12] << 8) + data[13]
    if self.B1 > 32767: self.B1 -= 65535
    self.B2 = (data[14] << 8) + data[15]
    if self.B2 > 32767: self.B2 -= 65535
    self.MB = (data[16] << 8) + data[17]
    if self.MB > 32767: self.MB -= 65535
    self.MC = (data[18] << 8) + data[19]
    if self.MC > 32767: self.MC -= 65535
    self.MD = (data[20] << 8) + data[21]
    if self.MD > 32767: self.MD -= 65535
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  def conv(self):
    '''
    conv()
    Запуск преобразования, считывание сырых данных с датчика,
    вычисление температуры и давления
    '''
    # Сначала температура
    self.bus.write_byte_data(self.addr, 0xF4, 0x2E)
    time.sleep(0.005) # Ждем не менее 4.5 мс
    data = self.bus.read_i2c_block_data(self.addr, 0xF6, 2)
    UT = (data[0] << 8) + data[1]
    # Потом давление (режим standart)
    self.bus.write_byte_data(self.addr, 0xF4, 0x74)
    time.sleep(0.008) # Ждем не менее 7.5 мс для режима standart
    data = self.bus.read_i2c_block_data(self.addr, 0xF6, 3)
    UP = ((data[0] << 16) + (data[1] << 8) + data[2])/128.0
    # Вычисляем реальную температуру
    X1 = (UT - self.AC6)*self.AC5/32768.0
    X2 = (self.MC*2048)/(X1 + self.MD)
    B5 = X1 + X2
    self.temp = (B5 + 8)/160.0
    # Вычисляем реальное давление
    B6 = B5 - 4000
    X1 = (self.B2*(B6*B6/4096.0))/2048.0
    X2 = self.AC2*B6/2048.0
    X3 = X1 + X2
    B3 = (((self.AC1*4 + X3)*2) + 2)/4.0
    X1 = self.AC3 * B6/8192.0
    X2 = (self.B1*(B6*B6/4096.0))/65536.0
    X3 = ((X1 + X2) + 2)/4.0
    B4 = self.AC4*(X3 + 32768)/32768.0
    B7 = ((UP - B3)*25000)
    if B7 < 0x80000000: p = (B7*2)/B4
    else: p = (B7/B4)*2
    X1 = (p/256.0)*(p/256.0)
    X1 = (X1*3038)/65536.0
    X2 = (-7357*p)/65536.0
    self.pres = p + (X1 + X2 + 3791)/16.0

  def update(self): # Затычка - перенести обработку сюда, убрать паузы в conv
    pass

  @property
  def val(self): # Текущее давление
    return self.pres*self.calibr[0]

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self._calibr

  @property
  def v(self): # Текущее давление с предварительным запуском conv и update
    self.conv()
    self.update()
    return self.val
#-------------------------------------------------------------------------------
class BMP280(object):
  ''' Класс датчика атмосферного давления BMP280 '''
  fc = 0x58
  def __init__(self, addr):
    '''
    __init__(self, addr)
    Конструктор. addr - адрес утройства на шине i2c.
    '''
    self.addr = addr
    self._calibr = [0.00750061683, 0, 5] # Pa -> mmHg
    self.bus = smbus.SMBus(1)
    self.temp = 0
    self.pres = 0
    # Читаем калибровочные данные
    data = self.bus.read_i2c_block_data(self.addr, 0x88, 24)
    self.dig_T1 = data[0] + (data[1] << 8)
    self.dig_T2 = data[2] + (data[3] << 8)
    if self.dig_T2 > 32767: self.dig_T2 -= 65536
    self.dig_T3 = data[4] + (data[5] << 8)
    if self.dig_T3 > 32767: self.dig_T3 -= 65536
    self.dig_P1 = data[6] + (data[7] << 8)
    self.dig_P2 = data[8] + (data[9] << 8)
    if self.dig_P2 > 32767: self.dig_P2 -= 65536
    self.dig_P3 = data[10] + (data[11] << 8)
    if self.dig_P3 > 32767: self.dig_P3 -= 65536
    self.dig_P4 = data[12] + (data[13] << 8)
    if self.dig_P4 > 32767: self.dig_P4 -= 65536
    self.dig_P5 = data[14] + (data[15] << 8)
    if self.dig_P5 > 32767: self.dig_P5 -= 65536
    self.dig_P6 = data[16] + (data[17] << 8)
    if self.dig_P6 > 32767: self.dig_P6 -= 65536
    self.dig_P7 = data[18] + (data[19] << 8)
    if self.dig_P7 > 32767: self.dig_P7 -= 65536
    self.dig_P8 = data[20] + (data[21] << 8)
    if self.dig_P8 > 32767: self.dig_P8 -= 65536
    self.dig_P9 = data[22] + (data[23] << 8)
    if self.dig_P9 > 32767: self.dig_P9 -= 65536
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  def conv(self):
    '''
    conv()
    Запуск преобразования, считывание сырых данных с датчика,
    вычисление температуры и давления
    '''
    # Основа кода преобразователя взята из примера:
    # http://www.pibits.net/code/raspberry-pi-and-bmp280-sensor-example.php
    # ...
    # Distributed with a free-will license.
    # Use it any way you want, profit or free, provided it fits in the licenses
    # of its associated works.
    # BMP280
    # This code is designed to work with the BMP280_I2CS I2C Mini Module
    # available from ControlEverything.com.
    # https://www.controleverything.com/content/Barometer?sku=BMP280_I2CSs#tabs-0-product_tabset-2
    # Select Control measurement register, 0xF4 0x27
    # Pressure and Temperature Oversampling rate = 1, Normal mode
    self.bus.write_byte_data(self.addr, 0xF4, 0x27)
    # Select Configuration register, 0xF5 0xA0(00) Stand_by time = 1000 ms
    self.bus.write_byte_data(self.addr, 0xF5, 0xA0)
#    time.sleep(0.5)
    time.sleep(0.05)
    # Read data back from 0xF7, 8 bytes: Pressure MSB, Pressure LSB,
    # Pressure xLSB, Temperature MSB, Temperature LSB, Temperature xLSB,
    # Humidity MSB, Humidity LSB
    data = self.bus.read_i2c_block_data(self.addr, 0xF7, 8)
    # Convert pressure and temperature data to 19-bits
    adc_p = ((data[0]*65536) + (data[1]*256) + (data[2] & 0xF0))/16
    adc_t = ((data[3]*65536) + (data[4]*256) + (data[5] & 0xF0))/16
    # Temperature offset calculations
    var1 = (adc_t/16384.0 - self.dig_T1/1024.0)*self.dig_T2
    var2 = (adc_t/131072.0 - self.dig_T1/8192.0)*\
           (adc_t/131072.0 - self.dig_T1/8192.0)*self.dig_T3
    t_fine = (var1 + var2)
    self.temp = (var1 + var2)/5120.0
    # Pressure offset calculations
    var1 = (t_fine/2.0) - 64000.0
    var2 = var1*var1*self.dig_P6/32768.0
    var2 = var2 + var1*self.dig_P5*2.0
    var2 = var2/4.0 + self.dig_P4*65536.0
    var1 = (self.dig_P3*var1*var1/524288.0 + self.dig_P2*var1)/524288.0
    var1 = (1.0 + var1/32768.0)*self.dig_P1
    p = 1048576.0 - adc_p
    p = (p - var2/4096.0)*6250.0/var1
    var1 = self.dig_P9*p*p/2147483648.0
    var2 = p*self.dig_P8/32768.0
    self.pres = p + (var1 + var2 + self.dig_P7)/16.0

  def update(self): # Затычка - перенести обработку сюда, убрать паузы в conv
    pass

  @property
  def val(self): # Текущее давление
    return self.pres*self.calibr[0]

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self._calibr

  @property
  def v(self): # Текущее давление с предварительным запуском conv и update
    self.conv()
    self.update()
    return self.val
#-------------------------------------------------------------------------------
class MS5611(object):
  ''' Класс датчика атмосферного давления MS5611 '''
  fc = 0x59
  def __init__(self, addr):
    '''
    __init__(self, addr)
    Конструктор. addr - адрес утройства на шине i2c.
    '''
    self.addr = addr
    self._calibr = [0.00750061683, 0, 5] # Pa -> mmHg
    self.bus = smbus.SMBus(1)
    self.temp = 0
    self.pres = 0
    self.bus.write_byte(self.addr, 0x1E) # Команда сброса
    time.sleep(0.5)
    # Читаем калибровочные данные
    data = self.bus.read_i2c_block_data(self.addr, 0xA2, 2)
    self.C1 = (data[0] << 8) + data[1]
    data = self.bus.read_i2c_block_data(self.addr, 0xA4, 2)
    self.C2 = (data[0] << 8) + data[1]
    data = self.bus.read_i2c_block_data(self.addr, 0xA6, 2)
    self.C3 = (data[0] << 8) + data[1]
    data = self.bus.read_i2c_block_data(self.addr, 0xA8, 2)
    self.C4 = (data[0] << 8) + data[1]
    data = self.bus.read_i2c_block_data(self.addr, 0xAA, 2)
    self.C5 = (data[0] << 8) + data[1]
    data = self.bus.read_i2c_block_data(self.addr, 0xAC, 2)
    self.C6 = (data[0] << 8) + data[1]
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  def conv(self):
    '''
    conv()
    Запуск преобразования. OSR = 1024, считывание сырых данных с датчика,
    вычисление температуры и давления
    '''
    self.bus.write_byte(self.addr, 0x44) # Запуск преобразования давления
    time.sleep(0.01)
    # Читаем и преобразуем три байта давления
    data = self.bus.read_i2c_block_data(self.addr, 0x00, 3)
    D1 = (data[0] << 16) + (data[1] << 8) + data[2]

    self.bus.write_byte(self.addr, 0x54) # Запуск преобразования температуры
    time.sleep(0.01)
    # Читаем и преобразуем три байта температуры
    data = self.bus.read_i2c_block_data(self.addr, 0x00, 3)
    D2 = (data[0] << 16) + (data[1] << 8) + data[2]
    # Вычисление температуры
    dT = D2 - self.C5*256
    TEMP = 2000 + (dT*self.C6)/8388608
    # Вычисление давления с температурной компенсацией
    OFF = self.C2*65536 + (self.C4*dT)/128
    SENS = self.C1*32768 + (self.C3*dT)/256

    if TEMP < 2000: # < 20°C
      T2 = (dT*dT)/2147483648
      OFF2 = (5*(TEMP - 2000)*(TEMP - 2000))/2
      SENS2 = (5*(TEMP - 2000) * (TEMP - 2000))/4
      if TEMP < -1500 :
        OFF2 = OFF2 + 7*(TEMP + 1500)*(TEMP + 1500)
        SENS2 = SENS2 + (11*(TEMP + 1500)*(TEMP + 1500))/2
    else:
      T2 = 0
      OFF2 = 0
      SENS2 = 0

    TEMP = TEMP - T2
    OFF = OFF - OFF2
    SENS = SENS - SENS2

    self.pres = (((D1*SENS)/2097152) - OFF)/32768.0
    self.temp = TEMP/100.0

  def update(self): # Затычка - перенести обработку сюда, убрать паузы в conv
    pass

  @property
  def val(self): # Текущее давление
    return self.pres*self.calibr[0]

  @property
  def calibr(self):
    '''Возвращает калибровочные данные'''
    return self._calibr

  @property
  def v(self): # Текущее давление с предварительным запуском conv и update
    self.conv()
    self.update()
    return self.val
#-------------------------------------------------------------------------------
class MG90S(I2C):
  ''' Привод сервомашинки MG90S
  '''
  fc = 0xA9
  def __init__(self, addr):
    '''
    __init__(self, addr)
    Конструктор. addr - адрес на шине i2c
    '''
    I2C.__init__(self, addr) # Инициализируем i2c в родительском классе
    self.calibr = [1.0, 0, 0] # Калибровочные данные - "затычка" для реле
    self.coefi = 1.0 # Обратная величина от коэффициента
    self._code = 0
    self._val = 0.0
    self.delay = 0 # Пошаговая пауза для замедления вращения в мс
    self.hkey = '' # hard-ключ (строка вида адрес_канал_тип)

  @property
  def val(self):
    return self._val

  @val.setter
  def val(self, value):
    buf = bytearray();
    buf.append(0x52)
    buf.append(value & 0xff)
    buf.append((value >> 8) & 0xff)
    buf.append(self.delay & 0xff)
    self.wd(buf)
    self._val = value
#-------------------------------------------------------------------------------
class TT(object):
  ''' Класс для работы с терминалом в неблокирующем режиме'''
  fc = 0xFE # Класс не связан с каким-нибудь устройством
  def __init__(self, timeout):
    self._timeout = timeout
    self._fd = sys.stdin.fileno()
    self._old_settings = new_settings = termios.tcgetattr(self._fd)
    self._poller = select.poll()
    self._poller.register(sys.stdin, select.POLLIN)
    tty.setcbreak(sys.stdin)
    fcntl.fcntl(self._fd, fcntl.F_SETFL, os.O_NONBLOCK)
  def __del__(self): # В деструкторе - возврат к старому терминалу
    termios.tcsetattr(self._fd, termios.TCSAFLUSH, self._old_settings)
  def kbhit(self): # Проверка нажата ли клавиша
    return self._poller.poll(self._timeout)
  def getch(self): # Возвращает введенный символ
    ch = sys.stdin.read(1)
    if ord(ch) == 27: # Esc-последовательность-выберем остаток в буфере
      sys.stdin.read()
    return ch
#-------------------------------------------------------------------------------
class Mode(object):
  ''' Класс режимов работы'''
  fc = 0xFE # Класс не связан с каким-нибудь устройством
  def __init__(self, name, init, cond):
    self.name = name
    self.init = init
    self.cond = cond
#-------------------------------------------------------------------------------
bus = smbus.SMBus(1)
'''Сформируем словарь всех классов данного модуля. Ключем является парамер
класса - fc (код семейства) '''
cls = {}
for cname, cobj in inspect.getmembers(sys.modules[__name__]):
  if inspect.isclass(cobj):
    cls[cobj.fc] = cobj
'''Сформируем словарь obj объектов-устройств, подключенных к шине i2c и каналам
хабов. Ключами являются hard-ключи (строки: \'адрес_канал_fc\'), а значениями -
списки из двух элементов - объект и его пользовательское имя (типа P0, T0, T1,
w0 и т.п.), которые будут определены позже'''
obj = {}
for addr in range(0x03, 0x78): # Пробегаем по адресам шины I2C
  try:
    bus.write_byte(addr, 0x33) # Запрос кода семейства
    time.sleep(0.1)
    if addr in (0x76, 0x77): # Датчики атм. давления (BMP180, BMP280, MS5611...)
      try: # Может быть это BMPxxx? Пробуем прочитать ID (код семейства)
        fc = bus.read_i2c_block_data(addr, 0xD0, 1)[0] # ID BMP-датчика
      except BaseException as error: # Увы... Что-то другое
        try: # А может это MS5611? - попробуем почитать регистры с калибр.
          data = bus.read_i2c_block_data(addr, 0xA2, 2)
          # Что-то приходит с этих регистров - предположим, что это MS5611
          fc = 0x59 # Назначим код семейства для MS5611
        except BaseException as error: # Неизвестное (пока) устройство
          print('\n! Устройство по адресу 0x%02X не распознано'%(addr))
          continue
    else: # Другие устройства на i2c - просто читаем код семейства (fc)
      fc = bus.read_byte(addr)
    o = cls[fc](addr) # Создаем объект обнаруженного устройства
    hkey = '%02X_%1d_%02X'%(addr, 8, fc) # hard-ключ устройства; 8 - нет каналов
    obj[hkey] = [o, None] # None - это имя устройства
    if fc == 0xA4: # Для контроллера PWM создаем виртуальный датчик расхода
      oo = cls[0xA8](o)
      hkey = '%02X_%1d_%02X'%(addr, 0, 0xA8) # hard-ключ устройства
      obj[hkey] = [oo, None]
    if fc == 0xA5: # Для хаба создаем объекты подключенных к нему датчиков
      for ch, chfc in enumerate(o.fcs):
        oo = cls[chfc](o, ch)
        hkey = '%02X_%1d_%02X'%(addr, ch, chfc) # hard-ключ устройства
        obj[hkey] = [oo, None]
  except IOError: # На этом адресе ничего нет - игнорируем
    pass
  except BaseException as error: # О других исключениях будем говорить ;)
    print('\n', str(error))

for hkey in sorted(obj.keys()):
  obj[hkey][0].hkey = hkey

import builtins
try:
  prnFlag = builtins.prnFlag
except AttributeError: # Флаг prnFlag не определен - установим его
  prnFlag = True

def hinfo(): # Вывод информации об обнаруженном оборудовании
  print('\nСписок обнаруженных устройств:')
  for hkey in sorted(obj.keys()):
    o = obj[hkey][0]
    if o.fc == 0xA5: # У хабов нет калибровочных данных
      print(obj[hkey][1], hkey, codes[o.fc][0])
    else:
      print(obj[hkey][1],   hkey, codes[o.fc][0], o.calibr)
'''Установка флага prnFlag в модуле builtins предполагает использование
библиотеки lite в "чистом виде", поэтому добавим дополнительные имена объектов,
согласно физическому смыслу устройств и выведем список этих объектов'''
if prnFlag:
  di = {} # Словарь для индексирования имен
  for key in codes.keys(): # Инициализируем словарь индексов
    let = codes[key][1] # Буква, соответствующая типу устройства
    di[let] = -1
  # Именуем объекты
  contr = [] # Список контроллеров
  for hkey in sorted(obj.keys()):
    o = obj[hkey][0]
    let = codes[o.fc][1] # Буква, соответствующая типу текущего устройства
    di[let] += 1 # Модифицируем индекс
    nam = let + str(di[let])
    obj[hkey][1] = nam
    if let >= 'a' and let <= 'z': # Маленькие буквы
      contr.append(nam)
    cod = nam + ' = obj[\'' + hkey + '\'][0]'
    exec(cod) # Создаем объект с именем nam

  hinfo() # Выведем на консоль информацию о найденном оборудовании
  
  cinit = 'def finit():\n' # Функция инициализации режима мониторинга
  for cnams in contr:
    cinit += '  ' + cnams + '.v = 0.0\n'
  exec(cinit)
  ccond = 'def fcond():\n  pass\n' # Функция проверок условий - пусто
  exec(ccond)
  m = Mode('Мониторинг', finit, fcond) # Создаем режим мониторинга
  modes = [m]
  mode = 0
  modes[mode].init()
#-------------------------------------------------------------------------------
''' Список выполненных и планируемых изменений
10.10.2019 Объявления классов варианта LITE выделено в отдельный модуль lite.py
           Это и есть данный файл.
           Добавлен класс MG90S для соответствующей сервомашинки
07.04.2020 v. 4.0.0.0
           1. В модуль добавлено именование объектов, соответственно физическому
              смыслу обнаруженных устройств
           2. Добавлена возможность обращения к объектам непосредственно по их
              именам (T0, w0 и т.п.)
           3. В класс RMS добавлен контроль актуальности данных (свойство v).
              То же самое сделано с классами DS18B20, MPX5010DP, Zcl и Zop.
           4. В классы датчиков атмосферного давления BMP180, BMP280 и MS5611
              добавлено св-во v с предварительным вызовом conv() и update()
           5. Добавлены свойства v в классы PDM, PWM и Vol в виде простых
              "затычек"
           6. Во все классы, соответствующие реальным объектам, добавлен
              атрибут hkey (hard-ключ)
10.04.2020 v. 4.0.0.1
           1. Добавлен список имен устройств-контроллеров (contr)
           2. Добавлен класс режимов Mode и 0-я мода - мониторинг
           3. Добавлен класс TT для неблокирующей работы терминала
14.04.2020 1. Значение атрибута fc классов Mode и TT заменено с 0x00 на 0xFE,
              чтобы не путался с нормально замкнутыми контактами
'''
#-------------------------------------------------------------------------------
